#!/bin/bash
########################################################################
# build_mpich.sh
#
# Build mpich and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_05_01.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib" && \
    ${trace} verify_in_filesystem -x "${@}" "${pkgdir}/bin/mpicxx"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

post_process() {
  # Fix rpaths in Fortran libraries.
  flibs=(libmpifort.dylib)
  if [ "${OS1}" = Darwin ]; then
    ${trace} pushd "${pkgdir}/lib"
    for flib in "${flibs[@]}"; do
      if [[ -f "${pkgdir}/lib/${flib}" ]]; then
        ${trace} otool -L "${flib}"
        ${trace} fix_macos_lib_symbols "${flib}" || return
      fi
    done
    ${trace} popd
  fi

  # Fix hard-wired paths everywhere we can.
  ${trace} find "${pkgdir}/"bin -type f -exec file \{\} \; | \
    ${trace} grep -Ee '^.*:.*\b(text|script)\b' | cut -d: -f1 | \
    while read file; do
    ${trace} fix_config_script "${file}" || return
  done

  # Remove libtool .la files, make pkg-config files relocatable and
  # verify.
  ${trace} rm_libtool_la && ${trace} fix_pc && ${trace} smoke_test_pc
}

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=mpich
origpkgver=v3_3_2
pkgver=${origpkgver}a # Append letters for local patches.
ssibuildshims_version=v2_06_00
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=${package}-${pkgdotver}
pkgtarfile=${pkgtarname}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e10,e14,e15,e17,e19,e20,c2,c5,c7} \
  --btype={debug,prof} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|TEST|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  || exit

##################
# We need to apply a patch that requires autogen.sh to be re-run, which
# requires a new-enough set of autotools.

patchtar="${patchdir}/${pkgtarname}-static-mpl.tar.bz2"
pre_autogen_patches=(hydra-Makefile-am.patch fix-mpl-configure.patch)

if [ -f "${patchtar}" ]; then
  ${trace} prepare_source "${ssi_args[@]}" "${patchtar}" || exit
else
  ssi_warn -i "Patch tarfile ${patchtar} is out of date or not present\n" \
    "for version ${pkgdotver}: attempting to generate with patch and autogen"
  [ -x gtar ] && tar=gtar || tar=tar
  tarver="$(${tar} --version | head -1)"
  if [[ "$tarver" == *GNU* ]]; then
    taropts=(--no-recursion --null -T- --newer-mtime "${sourcedir}/${pkgtarname}.stamp")
  elif [[ "$tarver" == bsdtar* ]]; then
    taropts=(-n --null -T- --newer-mtime-than "${sourcedir}/${pkgtarname}.stamp")
  else
    ssi_die "cannot generate autoconf patch archive: no GNU tar and don't understand tar version $tarver"
  fi
  ${trace} patch_source -a "${pkgtarname}" -P "${patchdir}" \
    "${pre_autogen_patches[@]}" || exit
  ${trace} cd "${pkgtarname}" || ssi_die "unable to cd to source $pkgtarname"
  ${trace} ./autogen.sh || \
    ssi_die "unable to run autogen.sh: missing or outdated autotools?"
  cd - >/dev/null
  ssi_info "creating patch archive ${patchtar}"
  # find is necessary to work around BSD tar's unfortunate tendency to
  # truncate at a directory not newer than the reference, even if files
  # lower down the hierarchy *are* newer than the reference.
  ${trace} find "${pkgtarname}" \! \( -type d \) -print0 | \
    ${trace} ${tar} jcf "$patchtar" "${taropts[@]}" || \
    ssi_die "unable to generate patch archive ${patchtar}"
  ssi_info "regenerating source archive in ${product_dir}"
  ${trace} make_source_code_tarball "${ssi_args[@]}" \
    "${product_dir}" ${package} ${pkgver} || \
    ssi_die "unable to regenerate source archive containing ${patchtar}"
fi

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

declare -a config_args cflg cxxflg ldflg mpich_cxxflg

config_args+=(--prefix="${pkgdir}"
  --enable-fortran=all --enable-cxx --enable-romio --enable-debuginfo
  --enable-versioning --enable-threads=runtime
  --disable-wrapper-rpath --enable-shared --disable-static
  --disable-libxml2 --enable-timer-type=clock_gettime --enable-g=debug
  --disable-dependency-tracking --disable-maintainer-mode
)

if [ "${OS1}" = Darwin ]; then
  config_args+=(--enable-timer-type=mach_absolute_time)
  ldflg+=(-L"${GCC_FQ_DIR}/lib" -lgfortran)
fi

case $btype in
  debug) config_args+=(
      --enable-error-checking=all --enable-error-messages=all
      --enable-timing=all --enable-mpit-pvars=all
      --disable-fast
    );;
  prof) config_args+=(--enable-mpit-pvars=none --enable-strict
      --enable-timing=none
      --enable-fast=ndebug,O3
    );;
  *) ssi_die -I "accepted unknown btype=$btype"
esac

case $cqual in
  c[27]) cxxstd=17; cc=clang; cxx=clang++;;
  e1[045]) cxxstd=14; cc=gcc; cxx=g++;;
  e1[79]|e20) cxxstd=17; cc=gcc; cxx=g++;;
  e21) cxxstd=20; cc=gcc; cxx=g++;;
  *) ssi_die -I "accepted unknown cqual=$cqual"
esac
fc=gfortran
mpich_cxxflg+=(-Wno-deprecated-declarations)
cflg+=(-D_LARGEFILE_SOURCE -D_LARGEFILE64_SOURCE -D_FILE_OFFSET_BITS=64)

##################
# Configure parallelism. *C*
ncores=$(ncores)
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*
if (( DEBUG_CONFIG )); then
  ${trace} find "${sourcedir}/${pkgtarname}" \
    -name configure | xargs sed -Ei'' -e '1 s&$& -x&' || \
    ssi_die "unable to update configure scripts for debug"
fi

${trace} env CC=${cc} CXX=${cxx} FC=${fc} \
  ${cxxflg:+"CXXFLAGS=${cxxflg[*]}"} \
  ${mpich_cxxflg:+"MPICH_CXXFLAGS=${mpich_cxxflg[*]}"} \
  ${cflg:+"CFLAGS=${cflg[*]}"} \
  ${ldflg:+"LDFLAGS=${ldflg[*]}"} \
  ${cxxstd:+stdver=c++${cxxstd}} \
  "${sourcedir}/${pkgtarname}/configure" \
  "${config_args[@]}" || ssi_die "configure failed"
##################

##################
# Build. *C*
${trace} make -j ${ncores} ${DEBUG_BUILD:+V=1} all || ssi_die "build failed"
##################

##################
# Test. *C*
if (( WANT_TESTS )); then
  ${trace} make -j ${ncores} ${DEBUG_TEST:+V=1} check || ssi_die "tests failed"
fi
##################

##################
# Install. *C*
${trace} make -j ${ncores} ${DEBUG_INSTALL:+V=1} install || ssi_die "install failed"
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.

# Use a function for nicer message / failure semantics.
ssi_info "post-processing for relocatability"
${trace} post_process || ssi_die "post-processing failed"
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
